""" GradCam Explainability"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from ..explain_utils import *
from .utils import PyTorchUtils, PyTorchGradCamModel


class GradCamPlusPlus:
    def __init__(self, model) -> None:

        self.model = model
        self.utils = PyTorchUtils()

    def _construct_gradcam_model(self, target_layer):
        model = PyTorchGradCamModel(self.model, target_layer)
        model.eval()
        return model

    def explain(
        self,
        input_array,
        explain_class=None,
        layer_index=None,
        heatmap_size=None,
        return_class=False,
    ):
        if layer_index is None:
            layer_index = self.utils.get_explainable_layers(self.model)[-2]

        gcmodel = self._construct_gradcam_model(layer_index)
        out, explaining_conv_layer_output = gcmodel(input_array)

        # If softmax not included, softmax the output. In PyTorch, softmax sometimes added to loss function.
        if out.max() > 1:
            out = torch.nn.functional.softmax(out)

        if explain_class is None:
            explain_class = out[0].argmax().item()

        loss = nn.CrossEntropyLoss()(out, torch.from_numpy(np.array([explain_class])))
        loss.backward()

        grads = gcmodel.get_act_grads().detach()

        score = out[0, explain_class]

        # Ref to original implementation : https://github.com/adityac94/Grad_CAM_plus_plus/blob/master/misc/utils.py
        # Second Grads
        grads = grads * out[0, explain_class].exp()
        grads_2 = grads**2
        grads_3 = grads**3

        global_sum = explaining_conv_layer_output.sum((2, 3))

        alpha_num = grads_2
        alpha_denom = 2 * grads_2 + (global_sum.reshape((1, -1, 1, 1)) * grads_3)
        alpha_denom = torch.where(
            alpha_denom != 0.0, alpha_denom, torch.ones(alpha_denom.shape)
        )
        alphas = alpha_num / alpha_denom

        weights = torch.clamp(grads, 0)

        # Convert to numpy to stick with original implementation
        weights = weights.detach().numpy()
        alphas = alphas.detach().numpy()

        alphas_thresholding = np.where(weights, alphas, 0.0)

        alpha_normalization_constant = alphas_thresholding.sum((0, 2, 3))
        alpha_normalization_constant_preprocessed = np.where(
            alpha_normalization_constant != 0.0,
            alpha_normalization_constant,
            np.ones(alpha_normalization_constant.shape),
        )

        alphas /= alpha_normalization_constant_preprocessed.reshape(1, -1, 1, 1)

        deep_linearization_weights = np.sum(
            (weights * alphas).reshape((-1, grads.shape[1])), axis=0
        )
        deep_linearization_weights = deep_linearization_weights.reshape(
            1, grads.shape[1], 1, 1
        )
        # Detach acts to numpy
        explaining_conv_layer_output = explaining_conv_layer_output.detach().numpy()

        heatmap = np.sum(
            deep_linearization_weights * explaining_conv_layer_output[0], axis=1
        )[0]

        # Passing through ReLU
        explanation = np.maximum(heatmap, 0)

        # Get image informations
        shape_list = list(input_array.shape)
        image_size = [i for i in shape_list if i > 4]
        if heatmap_size:
            image_size = heatmap_size
        channel = max([i for i in shape_list if i < 4])

        visualization = visualize(explanation, image_size, channel)

        if return_class:
            return visualization, explain_class
        return visualization



""" GradCam Explainability"""

import tensorflow as tf
import numpy as np

from .utils import *
from ..explain_utils import *


class GradCam:
    def __init__(self, model) -> None:
        self.model = model

    def explain(
        self,
        input_array,
        explain_class=None,
        layer_index=None,
        heatmap_size=None,
        separate=False,
        return_class=False,
    ):
        if separate:
            explaining_conv_layer_model, post_explain_model = separate_model(self.model)

            ## will be moved to keras utils
            with tf.GradientTape() as tape:
                inputs = input_array[np.newaxis, ...]
                explaining_conv_layer_output = explaining_conv_layer_model(inputs)
                tape.watch(explaining_conv_layer_output)
                preds = post_explain_model(explaining_conv_layer_output)
                if not explain_class:
                    explain_class = tf.argmax(preds[0])
                explain_class_channel = preds[:, explain_class]

            ## will be moved to keras utils
            grads = tape.gradient(top_class_channel, explaining_conv_layer_output)

        else:
            multioutput_model = create_multioutput_model(
                self.model, layer_index=layer_index
            )

            with tf.GradientTape() as tape:
                input_array = tf.cast(input_array, tf.float32)
                tape.watch(input_array)
                explaining_conv_layer_output, preds = multioutput_model(input_array)
                loss = preds[:, 0]
                if explain_class:
                    explain_class_channel = preds[:, explain_class]
                else:
                    top_pred_index = tf.argmax(preds[0])
                    explain_class_channel = preds[:, top_pred_index]

            grads = tape.gradient(explain_class_channel, explaining_conv_layer_output)

        pooled_grads = tf.reduce_mean(grads, axis=(0, 1, 2))

        explaining_conv_layer_output_0 = explaining_conv_layer_output.numpy()[0]
        pooled_grads = pooled_grads.numpy()
        for i in range(pooled_grads.shape[-1]):
            explaining_conv_layer_output_0[:, :, i] *= pooled_grads[i]

        # Average over all the filters to get a single 2D array
        explanation = np.mean(explaining_conv_layer_output_0, axis=-1)
        # Clip the values (equivalent to applying ReLU)
        # and then normalise the values

        # Get image informations
        shape_list = list(input_array.shape)
        image_size = [i for i in shape_list if i > 4]
        if heatmap_size:
            image_size = heatmap_size
        channel = max([i for i in shape_list if i < 4])

        visualization = visualize(explanation, image_size, channel)

        if return_class:
            return visualization, explain_class
        return visualization



import importlib
import os

from .pytorch import *
from .tensorflow import *

from .explain_utils import *


class Explain:
    def __init__(self, model):
        self.model = model
        self.model_framework = self._get_model_framework(model)

    def _get_model_framework(self, model):
        model_class_str = str(type(model))
        if "torch" in model_class_str:
            return "pytorch"
        elif "tensorflow" in model_class_str or "keras" in model_class_str:
            return "tensorflow"
        else:
            return None

    @staticmethod
    def get_pythonic_cwd():
        return os.getcwd().strip("/").replace("/", ".")

    def _import_method(self, method):
        explanation_module = importlib.import_module(
            f"why.explain.{self.model_framework}.{method.lower()}"
        )
        explain_class = getattr(explanation_module, method)
        return explain_class

    def explain(
        self,
        input_array,
        explain_class=None,
        method="GradCam",
        layer_index=None,
        heatmap_size=None,
        return_class=False,
    ):
        explanation_class = self._import_method(method)
        explanation_obj = explanation_class(self.model)

        return explanation_obj.explain(
            input_array=input_array,
            explain_class=explain_class,
            layer_index=layer_index,
            heatmap_size=heatmap_size,
            return_class=return_class,
        )

    def overlay_heatmap(
        self,
        original_image,
        heatmap,
        filename=None,
        image_size=None,
        alpha=0.5,
        colormap_name="RdYlBu",
        return_bytes=False,
    ):
        return overlay_heatmap_on_original_image(
            original_image,
            heatmap,
            filename=filename,
            image_size=image_size,
            alpha=alpha,
            colormap_name=colormap_name,
            return_bytes=return_bytes,
        )

    def annotate(
        self,
        input_array,
        image_size,
        threshold=0.7,
        explain_class=None,
        method="GradCam",
        layer_index=None,
    ):
        explanation_class = self._import_method(method)
        explanation_obj = explanation_class(self.model)

        heatmap, explain_class = explanation_obj.explain(
            input_array=input_array,
            explain_class=explain_class,
            layer_index=layer_index,
            heatmap_size=None,
            return_class=True,
        )

        resized_heatmap = resize_heatmap_wo_original_image(
            heatmap,
            filename=None,
            image_size=image_size,
            alpha=1,
            colormap_name="jet",
            return_bytes=False,
        )
        np_heatmap = np.array(resized_heatmap)
        mask = np_heatmap > int(threshold * 255)
        return create_polygon(mask[:, :, 0])

    def extract_area(
        self,
        input_array,
        original_image,
        threshold=0.7,
        explain_class=None,
        method="GradCam",
        layer_index=None,
    ):
        explanation_class = self._import_method(method)
        explanation_obj = explanation_class(self.model)

        heatmap, explain_class = explanation_obj.explain(
            input_array=input_array,
            explain_class=explain_class,
            layer_index=layer_index,
            heatmap_size=None,
            return_class=True,
        )

        resized_heatmap = resize_heatmap_wo_original_image(
            heatmap,
            filename=None,
            image_size=original_image.size,
            alpha=1,
            colormap_name="jet",
            return_bytes=False,
        )
        np_heatmap = np.array(resized_heatmap)
        mask = (np_heatmap > int(threshold * 255))[:, :, 0]

        np_original_image = np.array(original_image)
        extracted_area = []
        for i in range(np_original_image.shape[-1]):
            extracted_area.append(mask * np_original_image[:, :, i])
        extracted_area = np.stack(extracted_area, 2)
        return Image.fromarray(extracted_area)

    def resize_heatmap(
        self,
        heatmap,
        filename=None,
        image_size=None,
        alpha=0.5,
        colormap_name="RdYlBu",
        return_bytes=False,
    ):
        return resize_heatmap_wo_original_image(
            heatmap,
            filename=None,
            image_size=image_size,
            alpha=alpha,
            colormap_name=colormap_name,
            return_bytes=False,
        )



""" Keras CNN Layer utilities"""

import tensorflow as tf
import cv2
import numpy as np


def get_layers_type(model):
    """
    Returns model layer types dict of given Keras Model. Mainly used to find CNN layers which are appropriate to explain.
    """
    model_layers = {}
    for idx, layer in enumerate(model.layers):
        model_layers[idx] = {
            "layer_name": layer.name,
            "is_cnn": True if isinstance(layer, tf.keras.layers.Conv2D) else False,
        }
    return model_layers


def _get_explaining_layer(model, layer_index):
    # Use last CNN if index not given.

    model_layers = get_layers_type(model)
    if not layer_index:
        layer_index = [
            i
            for i in list(reversed(list(model_layers.keys())))
            if model_layers[i]["is_cnn"] == True
        ][0]
    return layer_index, model_layers


def separate_model(model, layer_index=None):
    """
    Separates model into explaining, and post-explain models.
    """

    layer_index, model_layers = _get_explaining_layer(model, layer_index=layer_index)

    explaining_conv_layer = model.get_layer(model_layers[layer_index]["layer_name"])
    explaining_conv_layer_model = tf.keras.Model(
        model.inputs, explaining_conv_layer.output
    )

    post_explain_input = tf.keras.Input(
        shape=explaining_conv_layer_model.output.shape[1:]
    )

    x = post_explain_input
    for layer_idx in list(model_layers.keys())[layer_index:]:
        x = model.get_layer(model_layers[layer_idx]["layer_name"])(x)

    post_explain_model = tf.keras.Model(post_explain_input, x)

    return explaining_conv_layer_model, post_explain_model


def create_multioutput_model(model, layer_index=None):
    """
    Creates model that gives explaining layers output, and model's normal output as well.
    This method is better than separating model into two. Separation and creation of two model is way problematic.
    """
    layer_index, model_layers = _get_explaining_layer(model, layer_index=layer_index)

    multioutput_model = tf.keras.Model(
        [model.inputs],
        [model.get_layer(model_layers[layer_index]["layer_name"]).output, model.output],
    )
    return multioutput_model



""" GradCam Explainability"""

import torch
import torch.nn as nn
import numpy as np

from ..explain_utils import *
from .utils import PyTorchUtils, PyTorchGradCamModel


class GradCam:
    def __init__(self, model) -> None:

        self.model = model
        self.utils = PyTorchUtils()

    def _construct_gradcam_model(self, target_layer):
        model = PyTorchGradCamModel(self.model, target_layer)
        model.eval()
        return model

    def explain(
        self,
        input_array,
        explain_class=None,
        layer_index=None,
        heatmap_size=None,
        return_class=False,
    ):
        if layer_index is None:
            layer_index = self.utils.get_explainable_layers(self.model)[-3]

        gcmodel = self._construct_gradcam_model(layer_index)
        out, acts = gcmodel(input_array)

        acts = acts.detach()

        if explain_class is None:
            explain_class = out[0].argmax().item()

        loss = nn.CrossEntropyLoss()(out, torch.from_numpy(np.array([explain_class])))
        loss.backward()

        grads = gcmodel.get_act_grads().detach()
        pooled_grads = torch.mean(grads, dim=[0, 2, 3]).detach().cpu()

        for i in range(acts.shape[1]):
            acts[:, i, :, :] *= pooled_grads[i]

        heatmap_j = torch.mean(acts, dim=1).squeeze()
        heatmap_j_max = heatmap_j.max(axis=0)[0]
        heatmap_j /= heatmap_j_max
        explanation = heatmap_j
        explanation = explanation.numpy()

        shape_list = list(input_array.shape)
        image_size = [i for i in shape_list if i > 4]
        channel = max([i for i in shape_list if i < 4])

        visualization = visualize(explanation, image_size, channel)
        if return_class:
            return visualization, explain_class
        return visualization



import io
import random
import copy
import string

from PIL import Image
import cv2
import numpy as np
import matplotlib.cm as mpl_color_map


def generate_random_name(N):
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(N)
    )


def visualize(map, image_size, channel):
    """
    Temporary visualization function
    """
    map = 255 * (map - map.min()) / (map.max() - map.min())
    map = cv2.resize(map, image_size)
    map = np.stack([map] * channel, -1)
    map = map.astype(np.uint8)
    return map


def array_handler(dl_array):
    if "torch" in str(type(dl_array)):
        return dl_array.numpy()
    else:
        return dl_array


def overlay_heatmap_on_original_image(
    original_image,
    heatmap,
    filename=None,
    image_size=None,
    alpha=0.5,
    colormap_name="RdYlBu",
    return_bytes=False,
):
    """
    Heatmap overlay
    """
    # Check if image is PIL or NumPy
    is_pil_or_np = True if "PIL" in str(type(original_image)) else False

    if image_size:
        if is_pil_or_np:
            original_image = original_image.resize(image_size)
        else:
            original_image = cv2.resize(original_image, image_size)
            original_image = Image.fromarray(original_image)

        heatmap = array_handler(heatmap)
        heatmap = np.array(Image.fromarray(heatmap).resize(image_size))
    else:
        image_size = (
            original_image.size
            if is_pil_or_np
            else [i for i in original_image.shape if i > 5]
        )

    color_map = mpl_color_map.get_cmap(colormap_name)

    # Change alpha channel in colormap to make sure original image is displayed
    heatmap_colored = color_map(heatmap[:, :, 0], alpha=alpha)
    heatmap_colored = (heatmap_colored * 255).astype(np.uint8)
    heatmap = copy.copy(heatmap_colored)

    heatmap = Image.fromarray((heatmap).astype(np.uint8))

    # Apply heatmap on image
    heatmap_on_image = Image.new("RGBA", original_image.size)
    heatmap_on_image = Image.alpha_composite(
        heatmap_on_image, original_image.convert("RGBA")
    )
    heatmap_on_image = Image.alpha_composite(heatmap_on_image, heatmap)

    if return_bytes:
        img_byte_arr = io.BytesIO()
        heatmap_on_image.save(img_byte_arr, format="PNG")
        return img_byte_arr.getvalue()
    else:
        if filename:
            is_extension = filename.lower().endswith(
                (".png", ".jpg", ".jpeg", ".tiff", ".bmp", ".gif")
            )
            filename = filename if is_extension else filename + ".png"
            heatmap_on_image.save(filename)
            return filename
        else:
            filename = generate_random_name(10) + ".png"
            heatmap_on_image.save(filename)
            return filename


def resize_heatmap_wo_original_image(
    heatmap,
    filename=None,
    image_size=None,
    alpha=0.5,
    colormap_name="RdYlBu",
    return_bytes=False,
):
    """
    Heatmap overlay
    """

    if image_size:
        heatmap = np.array(Image.fromarray(heatmap).resize(image_size))

    color_map = mpl_color_map.get_cmap(colormap_name)

    # Change alpha channel in colormap to make sure original image is displayed
    heatmap_colored = color_map(heatmap[:, :, 0], alpha=alpha)
    heatmap_colored = (heatmap_colored * 255).astype(np.uint8)
    heatmap = copy.copy(heatmap_colored)

    heatmap = Image.fromarray((heatmap).astype(np.uint8))

    if return_bytes:
        img_byte_arr = io.BytesIO()
        heatmap.save(img_byte_arr, format="PNG")
        return img_byte_arr.getvalue()
    else:
        if filename:
            is_extension = filename.lower().endswith(
                (".png", ".jpg", ".jpeg", ".tiff", ".bmp", ".gif")
            )
            filename = filename if is_extension else filename + ".png"
            heatmap.save(filename)
            return filename
        else:
            return heatmap


def create_polygon(mask):

    contours, _ = cv2.findContours(
        mask.astype(np.uint8), cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE
    )
    polygons = []
    for object in contours:
        for point in object:
            polygons.append({"x": int(point[0][0]), "y": int(point[0][1])})
        break
    return polygons



""" Keras CNN Layer utilities"""
import torch
import torch.nn as nn


class PyTorchGradCamModel(nn.Module):
    def __init__(self, model, layer_name):
        super().__init__()
        self.gradients = None
        self.tensorhook = []
        self.layerhook = []
        self.selected_out = None
        self.model = model

        self.layerhook.append(
            getattr(model, layer_name).register_forward_hook(self.forward_hook())
        )

        for p in self.model.parameters():
            p.requires_grad = True

    def activations_hook(self, grad):
        self.gradients = grad

    def get_act_grads(self):
        return self.gradients

    def forward_hook(self):
        def hook(module, inp, out):
            self.selected_out = out
            self.tensorhook.append(out.register_hook(self.activations_hook))

        return hook

    def forward(self, x):
        out = self.model(x)
        return out, self.selected_out


class PyTorchUtils:
    def __init__(
        self,
    ):
        pass

    def get_explainable_layers(self, model):
        """
        xai_layers = []

        for name, modules in model.named_modules():
            if isinstance(modules, torch.nn.Sequential):
                xai_layers.append(name)

        TO DO: Work on blocked models
        """
        return list(model.__dict__["_modules"].keys())



from setuptools import setup, find_packages


packages = find_packages()


packages.append("why")


setup(
    name="Why-XAI",
    version="0.1.0",
    description="Framework-agnostic CNN XAI Library",
    url="https://github.com/mburaksayici",
    packages=packages,
    author="Mehmet Burak Sayıcı",
    author_email="mburaksayici@gmail.com",
    license="BSD 2-clause",
    install_requires=["tensorflow", "numpy", "torch", "torchvision", "opencv-python", "matplotlib"],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



""" GradCam Explainability"""

import tensorflow as tf
import numpy as np

from .utils import *
from ..explain_utils import *


class GradCamPlusPlus:
    """
    Gradcam++, referenced from : https://github.com/adityac94/Grad_CAM_plus_plus/blob/master/misc/utils.py
    Lots of different implementations with significant differences from paper/other implementations. Sticking to original one.
    """

    def __init__(self, model) -> None:
        self.model = model

    def explain(
        self,
        input_array,
        explain_class=None,
        layer_index=None,
        heatmap_size=None,
        separate=False,
        return_class=False,
    ):
        if separate:
            explaining_conv_layer_model, post_explain_model = separate_model(self.model)

            ## will be moved to keras utils
            with tf.GradientTape() as tape:
                inputs = input_array[np.newaxis, ...]
                explaining_conv_layer_output = explaining_conv_layer_model(inputs)
                tape.watch(explaining_conv_layer_output)
                preds = post_explain_model(explaining_conv_layer_output)
                if not explain_class:
                    explain_class = tf.argmax(preds[0])
                explain_class_channel = preds[:, explain_class]

            ## will be moved to keras utils
            grads = tape.gradient(top_class_channel, explaining_conv_layer_output)

        else:
            multioutput_model = create_multioutput_model(
                self.model, layer_index=layer_index
            )

            with tf.GradientTape() as tape:
                input_array = tf.cast(input_array, tf.float32)
                tape.watch(input_array)
                explaining_conv_layer_output, preds = multioutput_model(input_array)
                if explain_class:
                    explain_class_channel = preds[:, explain_class]
                else:
                    top_pred_index = tf.argmax(preds[0])
                    explain_class_channel = preds[:, top_pred_index]

            grads = tape.gradient(explain_class_channel, explaining_conv_layer_output)

        # Grads
        grads = grads * tf.exp(explain_class_channel)
        grads_2 = grads**2
        grads_3 = grads_2 * grads

        explaining_conv_layer_output = explaining_conv_layer_output.numpy()

        global_sum = np.sum(
            explaining_conv_layer_output[0].reshape((-1, grads[0].shape[2])), axis=0
        )

        alpha_num = grads_2[0]
        alpha_denom = grads_2[0] * 2.0 + grads_3[0] * global_sum.reshape(
            (1, 1, grads[0].shape[2])
        )
        alpha_denom = np.where(
            alpha_denom != 0.0, alpha_denom, np.ones(alpha_denom.shape)
        )
        alphas = alpha_num / alpha_denom

        weights = np.maximum(grads[0], 0.0)

        alphas_thresholding = np.where(weights, alphas, 0.0)

        alpha_normalization_constant = np.sum(
            np.sum(alphas_thresholding, axis=0), axis=0
        )
        alpha_normalization_constant_processed = np.where(
            alpha_normalization_constant != 0.0,
            alpha_normalization_constant,
            np.ones(alpha_normalization_constant.shape),
        )

        alphas /= alpha_normalization_constant_processed.reshape(
            (1, 1, grads[0].shape[2])
        )
        alphas = alphas.numpy()

        deep_linearization_weights = np.sum(
            (weights * alphas).reshape((-1, grads[0].shape[2])), axis=0
        )
        # print deep_linearization_weights
        heatmap = np.sum(
            deep_linearization_weights * explaining_conv_layer_output[0], axis=2
        )

        # Passing through ReLU
        explanation = np.maximum(heatmap, 0)

        # Get image informations
        shape_list = list(input_array.shape)
        image_size = [i for i in shape_list if i > 4]
        if heatmap_size:
            image_size = heatmap_size
        channel = max([i for i in shape_list if i < 4])

        visualization = visualize(explanation, image_size, channel)

        if return_class:
            return visualization, explain_class
        return visualization
